# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous useful stuff."""


def _common_path_and_rest(l1, l2, common=[]):
    # From http://code.activestate.com/recipes/208993/
    if len(l1) < 1:
        return (common, l1, l2)
    if len(l2) < 1:
        return (common, l1, l2)
    if l1[0] != l2[0]:
        return (common, l1, l2)
    return _common_path_and_rest(
        l1[1:],
        l2[1:],
        common + [
            l1[0:1]  # return a byte string in python 3 unlike l1[0] that
                     # would return an integer.
        ]
    )


def common_path(path1, path2):
    """Find the common bit of 2 paths."""
    return b''.join(_common_path_and_rest(path1, path2)[0])


def common_directory(paths):
    """Find the deepest common directory of a list of paths.

    :return: if no paths are provided, None is returned;
      if there is no common directory, '' is returned;
      otherwise the common directory with a trailing / is returned.
    """
    import posixpath

    def get_dir_with_slash(path):
        if path == b'' or path.endswith(b'/'):
            return path
        else:
            dirname, basename = posixpath.split(path)
            if dirname == b'':
                return dirname
            else:
                return dirname + b'/'

    if not paths:
        return None
    elif len(paths) == 1:
        return get_dir_with_slash(paths[0])
    else:
        common = common_path(paths[0], paths[1])
        for path in paths[2:]:
            common = common_path(common, path)
        return get_dir_with_slash(common)


def is_inside(directory, fname):
    """True if fname is inside directory.

    The parameters should typically be passed to osutils.normpath first, so
    that . and .. and repeated slashes are eliminated, and the separators
    are canonical for the platform.

    The empty string as a dir name is taken as top-of-tree and matches
    everything.
    """
    # XXX: Most callers of this can actually do something smarter by
    # looking at the inventory
    if directory == fname:
        return True

    if directory == b'':
        return True

    if not directory.endswith(b'/'):
        directory += b'/'

    return fname.startswith(directory)


def is_inside_any(dir_list, fname):
    """True if fname is inside any of given dirs."""
    for dirname in dir_list:
        if is_inside(dirname, fname):
            return True
    return False


def utf8_bytes_string(s):
    """Convert a string to a bytes string (if necessary, encode in utf8)"""
    if isinstance(s, str):
        return bytes(s, encoding='utf8')
    else:
        return s


class newobject(object):
    """
    A magical object class that provides Python 2 compatibility methods::
        next
        __unicode__
        __nonzero__

    Subclasses of this class can merely define the Python 3 methods (__next__,
    __str__, and __bool__).

    This is a copy/paste of the future.types.newobject class of the future
    package.
    """
    def next(self):
        if hasattr(self, '__next__'):
            return type(self).__next__(self)
        raise TypeError('newobject is not an iterator')

    def __unicode__(self):
        # All subclasses of the builtin object should have __str__ defined.
        # Note that old-style classes do not have __str__ defined.
        if hasattr(self, '__str__'):
            s = type(self).__str__(self)
        else:
            s = str(self)
        if isinstance(s, unicode):  # noqa: F821
            return s
        else:
            return s.decode('utf-8')

    def __nonzero__(self):
        if hasattr(self, '__bool__'):
            return type(self).__bool__(self)
        # object has no __nonzero__ method
        return True

    # Are these ever needed?
    # def __div__(self):
    #     return self.__truediv__()

    # def __idiv__(self, other):
    #     return self.__itruediv__(other)

    def __long__(self):
        if not hasattr(self, '__int__'):
            return NotImplemented
        return self.__int__()  # not type(self).__int__(self)

    # def __new__(cls, *args, **kwargs):
    #     """
    #     dict() -> new empty dictionary
    #     dict(mapping) -> new dictionary initialized from a mapping object's
    #         (key, value) pairs
    #     dict(iterable) -> new dictionary initialized as if via:
    #         d = {}
    #         for k, v in iterable:
    #             d[k] = v
    #     dict(**kwargs) -> new dictionary initialized with the name=value
    #         pairs in the keyword argument list.
    #         For example:  dict(one=1, two=2)
    #     """

    #     if len(args) == 0:
    #         return super(newdict, cls).__new__(cls)
    #     elif type(args[0]) == newdict:
    #         return args[0]
    #     else:
    #         value = args[0]
    #     return super(newdict, cls).__new__(cls, value)

    def __native__(self):
        """
        Hook for the future.utils.native() function
        """
        return object(self)


def binary_stream(stream):
    """Ensure a stream is binary on Windows.

    :return: the stream
    """
    try:
        import os
        if os.name == 'nt':
            fileno = getattr(stream, 'fileno', None)
            if fileno:
                no = fileno()
                if no >= 0:     # -1 means we're working as subprocess
                    import msvcrt
                    msvcrt.setmode(no, os.O_BINARY)
    except ImportError:
        pass
    return stream


def invert_dictset(d):
    """Invert a dict with keys matching a set of values, turned into lists."""
    # Based on recipe from ASPN
    result = {}
    for k, c in d.items():
        for v in c:
            keys = result.setdefault(v, [])
            keys.append(k)
    return result


def invert_dict(d):
    """Invert a dictionary with keys matching each value turned into a list."""
    # Based on recipe from ASPN
    result = {}
    for k, v in d.items():
        keys = result.setdefault(v, [])
        keys.append(k)
    return result


def defines_to_dict(defines):
    """Convert a list of definition strings to a dictionary."""
    if defines is None:
        return None
    result = {}
    for define in defines:
        kv = define.split('=', 1)
        if len(kv) == 1:
            result[define.strip()] = 1
        else:
            result[kv[0].strip()] = kv[1].strip()
    return result


def get_source_stream(source):
    if source == '-' or source is None:
        import sys
        stream = binary_stream(sys.stdin)
    elif source.endswith('.gz'):
        import gzip
        stream = gzip.open(source, "rb")
    else:
        stream = open(source, "rb")
    return stream



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test how Commands are displayed"""

from unittest import TestCase

from fastimport.helpers import (
    utf8_bytes_string,
    )

from fastimport import (
    commands,
    )


class TestBlobDisplay(TestCase):

    def test_blob(self):
        c = commands.BlobCommand(b"1", b"hello world")
        self.assertEqual(b"blob\nmark :1\ndata 11\nhello world", bytes(c))

    def test_blob_no_mark(self):
        c = commands.BlobCommand(None, b"hello world")
        self.assertEqual(b"blob\ndata 11\nhello world", bytes(c))


class TestCheckpointDisplay(TestCase):

    def test_checkpoint(self):
        c = commands.CheckpointCommand()
        self.assertEqual(b'checkpoint', bytes(c))


class TestCommitDisplay(TestCase):

    def test_commit(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b"refs/heads/master", b"bbb", None, committer,
            b"release v1.0", b":aaa", None, None)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa",
            bytes(c))

    def test_commit_unicode_committer(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        name = u'\u013d\xf3r\xe9m \xcdp\u0161\xfam'

        commit_utf8 = utf8_bytes_string(
            u"commit refs/heads/master\n"
            u"mark :bbb\n"
            u"committer %s <test@example.com> 1234567890 -0600\n"
            u"data 12\n"
            u"release v1.0\n"
            u"from :aaa" % (name,)
        )

        committer = (name, b'test@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', None, committer,
            b'release v1.0', b':aaa', None, None)

        self.assertEqual(commit_utf8, bytes(c))

    def test_commit_no_mark(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b'refs/heads/master', None, None, committer,
            b'release v1.0', b':aaa', None, None)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa",
            bytes(c))

    def test_commit_no_from(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b"refs/heads/master", b"bbb", None, committer,
            b"release v1.0", None, None, None)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0",
            bytes(c))

    def test_commit_with_author(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        author = (b'Sue Wong', b'sue@example.com', 1234565432, -6 * 3600)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', author,
            committer, b'release v1.0', b':aaa', None, None)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"author Sue Wong <sue@example.com> 1234565432 -0600\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa",
            bytes(c))

    def test_commit_with_merges(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
                b"refs/heads/master", b"ddd", None, committer,
                b'release v1.0', b":aaa", [b':bbb', b':ccc'], None)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :ddd\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa\n"
            b"merge :bbb\n"
            b"merge :ccc",
            bytes(c))

    def test_commit_with_filecommands(self):
        file_cmds = iter([
            commands.FileDeleteCommand(b'readme.txt'),
            commands.FileModifyCommand(
                b'NEWS', 0o100644, None, b'blah blah blah'),
            ])
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', None, committer,
            b'release v1.0', b':aaa', None, file_cmds)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa\n"
            b"D readme.txt\n"
            b"M 644 inline NEWS\n"
            b"data 14\n"
            b"blah blah blah",
            bytes(c))

    def test_commit_with_more_authors(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        author = (b'Sue Wong', b'sue@example.com', 1234565432, -6 * 3600)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        more_authors = [
            (b'Al Smith', b'al@example.com', 1234565432, -6 * 3600),
            (b'Bill Jones', b'bill@example.com', 1234565432, -6 * 3600),
        ]
        c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', author,
            committer, b'release v1.0', b':aaa', None, None,
            more_authors=more_authors)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"author Sue Wong <sue@example.com> 1234565432 -0600\n"
            b"author Al Smith <al@example.com> 1234565432 -0600\n"
            b"author Bill Jones <bill@example.com> 1234565432 -0600\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa",
            bytes(c))

    def test_commit_with_properties(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        properties = {
            u'greeting':  u'hello',
            u'planet':    u'world',
            }
        c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', None,
            committer, b'release v1.0', b':aaa', None, None,
            properties=properties)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :bbb\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa\n"
            b"property greeting 5 hello\n"
            b"property planet 5 world",
            bytes(c))

    def test_commit_with_int_mark(self):
        # user tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        properties = {
            u'greeting':  u'hello',
            u'planet':    u'world',
            }
        c = commands.CommitCommand(
            b'refs/heads/master', 123, None,
            committer, b'release v1.0', b':aaa', None, None,
            properties=properties)
        self.assertEqual(
            b"commit refs/heads/master\n"
            b"mark :123\n"
            b"committer Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 12\n"
            b"release v1.0\n"
            b"from :aaa\n"
            b"property greeting 5 hello\n"
            b"property planet 5 world",
            bytes(c))


class TestCommitCopy(TestCase):

    def setUp(self):
        super(TestCommitCopy, self).setUp()
        file_cmds = iter([
            commands.FileDeleteCommand(b'readme.txt'),
            commands.FileModifyCommand(
                b'NEWS', 0o100644, None, b'blah blah blah'),
        ])

        committer = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        self.c = commands.CommitCommand(
            b'refs/heads/master', b'bbb', None, committer,
            b'release v1.0', b':aaa', None, file_cmds)

    def test_simple_copy(self):
        c2 = self.c.copy()

        self.assertFalse(self.c is c2)
        self.assertEqual(bytes(self.c), bytes(c2))

    def test_replace_attr(self):
        c2 = self.c.copy(mark=b'ccc')
        self.assertEqual(
            bytes(self.c).replace(b'mark :bbb', b'mark :ccc'),
            bytes(c2)
        )

    def test_invalid_attribute(self):
        self.assertRaises(TypeError, self.c.copy, invalid=True)


class TestFeatureDisplay(TestCase):

    def test_feature(self):
        c = commands.FeatureCommand(b"dwim")
        self.assertEqual(b"feature dwim", bytes(c))

    def test_feature_with_value(self):
        c = commands.FeatureCommand(b"dwim", b"please")
        self.assertEqual(b"feature dwim=please", bytes(c))


class TestProgressDisplay(TestCase):

    def test_progress(self):
        c = commands.ProgressCommand(b"doing foo")
        self.assertEqual(b"progress doing foo", bytes(c))


class TestResetDisplay(TestCase):

    def test_reset(self):
        c = commands.ResetCommand(b"refs/tags/v1.0", b":xxx")
        self.assertEqual(b"reset refs/tags/v1.0\nfrom :xxx\n", bytes(c))

    def test_reset_no_from(self):
        c = commands.ResetCommand(b'refs/remotes/origin/master', None)
        self.assertEqual(b'reset refs/remotes/origin/master', bytes(c))


class TestTagDisplay(TestCase):

    def test_tag(self):
        # tagger tuple is (name, email, secs-since-epoch, secs-offset-from-utc)
        tagger = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.TagCommand(
            b'refs/tags/v1.0', b':xxx', tagger, b'create v1.0')
        self.assertEqual(
            b"tag refs/tags/v1.0\n"
            b"from :xxx\n"
            b"tagger Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 11\n"
            b"create v1.0",
            bytes(c))

    def test_tag_no_from(self):
        tagger = (b'Joe Wong', b'joe@example.com', 1234567890, -6 * 3600)
        c = commands.TagCommand(
            b'refs/tags/v1.0', None, tagger, b'create v1.0')
        self.assertEqual(
            b"tag refs/tags/v1.0\n"
            b"tagger Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 11\n"
            b"create v1.0",
            bytes(c))


class TestFileModifyDisplay(TestCase):

    def test_filemodify_file(self):
        c = commands.FileModifyCommand(b'foo/bar', 0o100644, b':23', None)
        self.assertEqual(b'M 644 :23 foo/bar', bytes(c))

    def test_filemodify_file_executable(self):
        c = commands.FileModifyCommand(b'foo/bar', 0o100755, b':23', None)
        self.assertEqual(b'M 755 :23 foo/bar', bytes(c))

    def test_filemodify_file_internal(self):
        c = commands.FileModifyCommand(
            b'foo/bar', 0o100644, None, b'hello world')
        self.assertEqual(
            b'M 644 inline foo/bar\ndata 11\nhello world', bytes(c))

    def test_filemodify_symlink(self):
        c = commands.FileModifyCommand(b'foo/bar', 0o120000, None, b'baz')
        self.assertEqual(
            b'M 120000 inline foo/bar\ndata 3\nbaz', bytes(c))

    def test_filemodify_treeref(self):
        c = commands.FileModifyCommand(
            b'tree-info', 0o160000, b'revision-id-info', None)
        self.assertEqual(
            b'M 160000 revision-id-info tree-info', bytes(c))


class TestFileDeleteDisplay(TestCase):

    def test_filedelete(self):
        c = commands.FileDeleteCommand(b'foo/bar')
        self.assertEqual(b'D foo/bar', bytes(c))


class TestFileCopyDisplay(TestCase):

    def test_filecopy(self):
        c = commands.FileCopyCommand(b'foo/bar', b'foo/baz')
        self.assertEqual(b'C foo/bar foo/baz', bytes(c))

    def test_filecopy_quoted(self):
        # Check the first path is quoted if it contains spaces
        c = commands.FileCopyCommand(b'foo/b a r', b'foo/b a z')
        self.assertEqual(b'C "foo/b a r" foo/b a z', bytes(c))


class TestFileRenameDisplay(TestCase):

    def test_filerename(self):
        c = commands.FileRenameCommand(b'foo/bar', b'foo/baz')
        self.assertEqual(b'R foo/bar foo/baz', bytes(c))

    def test_filerename_quoted(self):
        # Check the first path is quoted if it contains spaces
        c = commands.FileRenameCommand(b'foo/b a r', b'foo/b a z')
        self.assertEqual(b'R "foo/b a r" foo/b a z', bytes(c))


class TestFileDeleteAllDisplay(TestCase):

    def test_filedeleteall(self):
        c = commands.FileDeleteAllCommand()
        self.assertEqual(b'deleteall', bytes(c))


class TestNotesDisplay(TestCase):

    def test_noteonly(self):
        c = commands.NoteModifyCommand(b'foo', b'A basic note')
        self.assertEqual(
            b'N inline :foo\ndata 12\nA basic note', bytes(c))

    def test_notecommit(self):
        committer = (b'Ed Mund', b'ed@example.org', 1234565432, 0)

        commits = [
            commands.CommitCommand(
                ref=b'refs/heads/master',
                mark=b'1',
                author=committer,
                committer=committer,
                message=b'test\n',
                from_=None,
                merges=[],
                file_iter=[
                    commands.FileModifyCommand(b'bar', 0o100644, None, b'')
                ]),
            commands.CommitCommand(
                ref=b'refs/notes/commits',
                mark=None,
                author=None,
                committer=committer,
                message=b"Notes added by 'git notes add'\n",
                from_=None,
                merges=[],
                file_iter=[
                    commands.NoteModifyCommand(b'1', b'Test note\n')
                ]),
            commands.CommitCommand(
                ref=b'refs/notes/test',
                mark=None,
                author=None,
                committer=committer,
                message=b"Notes added by 'git notes add'\n",
                from_=None,
                merges=[],
                file_iter=[
                    commands.NoteModifyCommand(b'1', b'Test test\n')
                ])
        ]

        self.assertEqual(
            b"""commit refs/heads/master
mark :1
author Ed Mund <ed@example.org> 1234565432 +0000
committer Ed Mund <ed@example.org> 1234565432 +0000
data 5
test

M 644 inline bar
data 0
commit refs/notes/commits
committer Ed Mund <ed@example.org> 1234565432 +0000
data 31
Notes added by 'git notes add'

N inline :1
data 10
Test note
commit refs/notes/test
committer Ed Mund <ed@example.org> 1234565432 +0000
data 31
Notes added by 'git notes add'

N inline :1
data 10
Test test
""", b''.join([bytes(s) for s in commits]))


class TestPathChecking(TestCase):

    def test_filemodify_path_checking(self):
        self.assertRaises(
            ValueError, commands.FileModifyCommand, b'',
            0o100644, None, b'text')
        self.assertRaises(
            ValueError, commands.FileModifyCommand, None,
            0o100644, None, b'text')

    def test_filedelete_path_checking(self):
        self.assertRaises(ValueError, commands.FileDeleteCommand, b'')
        self.assertRaises(ValueError, commands.FileDeleteCommand, None)

    def test_filerename_path_checking(self):
        self.assertRaises(ValueError, commands.FileRenameCommand, b'', b'foo')
        self.assertRaises(ValueError, commands.FileRenameCommand, None, b'foo')
        self.assertRaises(ValueError, commands.FileRenameCommand, b'foo', b'')
        self.assertRaises(ValueError, commands.FileRenameCommand, b'foo', None)

    def test_filecopy_path_checking(self):
        self.assertRaises(ValueError, commands.FileCopyCommand, b'', b'foo')
        self.assertRaises(ValueError, commands.FileCopyCommand, None, b'foo')
        self.assertRaises(ValueError, commands.FileCopyCommand, b'foo', b'')
        self.assertRaises(ValueError, commands.FileCopyCommand, b'foo', None)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Processor for fast-import commands.

This module provides the skeleton of a fast-import backend.
To import from a fast-import stream to your version-control system:

 - derive a class from the abstract ImportProcessor class and
   implement the *_helper methods.

 - parse a fast-import stream into a sequence of commands, for example
   using the helpers from the parser module.

 - pass that command sequence to the process method of your processor.

See git-fast-import.1 for the meaning of each command and the
processors package for examples.
"""
import sys
import time

from . import errors
from .helpers import newobject as object


class ImportProcessor(object):
    """Base class for fast-import stream processors.

    Subclasses should override the pre_*, post_* and *_handler
    methods as appropriate.
    """

    known_params = []

    def __init__(self, params=None, verbose=False, outf=None):
        if outf is None:
            self.outf = sys.stdout
        else:
            self.outf = outf
        self.verbose = verbose
        if params is None:
            self.params = {}
        else:
            self.params = params
            self.validate_parameters()

        # Handlers can set this to request exiting cleanly without
        # iterating through the remaining commands
        self.finished = False

    def validate_parameters(self):
        """Validate that the parameters are correctly specified."""
        for p in self.params:
            if p not in self.known_params:
                raise errors.UnknownParameter(p, self.known_params)

    def process(self, command_iter):
        """Import data into Bazaar by processing a stream of commands.

        :param command_iter: an iterator providing commands
        """
        self._process(command_iter)

    def _process(self, command_iter):
        self.pre_process()
        for cmd in command_iter():
            try:
                name = (cmd.name + b'_handler').decode('utf8')
                handler = getattr(self.__class__, name)
            except KeyError:
                raise errors.MissingHandler(cmd.name)
            else:
                self.pre_handler(cmd)
                handler(self, cmd)
                self.post_handler(cmd)
            if self.finished:
                break
        self.post_process()

    def warning(self, msg, *args):
        """Output a warning but timestamp it."""
        pass

    def debug(self, mgs, *args):
        """Output a debug message."""
        pass

    def _time_of_day(self):
        """Time of day as a string."""
        # Note: this is a separate method so tests can patch in a fixed value
        return time.strftime("%H:%M:%S")

    def pre_process(self):
        """Hook for logic at start of processing."""
        pass

    def post_process(self):
        """Hook for logic at end of processing."""
        pass

    def pre_handler(self, cmd):
        """Hook for logic before each handler starts."""
        pass

    def post_handler(self, cmd):
        """Hook for logic after each handler finishes."""
        pass

    def progress_handler(self, cmd):
        """Process a ProgressCommand."""
        raise NotImplementedError(self.progress_handler)

    def blob_handler(self, cmd):
        """Process a BlobCommand."""
        raise NotImplementedError(self.blob_handler)

    def checkpoint_handler(self, cmd):
        """Process a CheckpointCommand."""
        raise NotImplementedError(self.checkpoint_handler)

    def commit_handler(self, cmd):
        """Process a CommitCommand."""
        raise NotImplementedError(self.commit_handler)

    def reset_handler(self, cmd):
        """Process a ResetCommand."""
        raise NotImplementedError(self.reset_handler)

    def tag_handler(self, cmd):
        """Process a TagCommand."""
        raise NotImplementedError(self.tag_handler)

    def feature_handler(self, cmd):
        """Process a FeatureCommand."""
        raise NotImplementedError(self.feature_handler)


class CommitHandler(object):
    """Base class for commit handling.

    Subclasses should override the pre_*, post_* and *_handler
    methods as appropriate.
    """

    def __init__(self, command):
        self.command = command

    def process(self):
        self.pre_process_files()
        for fc in self.command.iter_files():
            try:
                name = (fc.name[4:] + b'_handler').decode('utf8')
                handler = getattr(self.__class__, name)
            except KeyError:
                raise errors.MissingHandler(fc.name)
            else:
                handler(self, fc)
        self.post_process_files()

    def warning(self, msg, *args):
        """Output a warning but add context."""
        pass

    def pre_process_files(self):
        """Prepare for committing."""
        pass

    def post_process_files(self):
        """Save the revision."""
        pass

    def modify_handler(self, filecmd):
        """Handle a filemodify command."""
        raise NotImplementedError(self.modify_handler)

    def delete_handler(self, filecmd):
        """Handle a filedelete command."""
        raise NotImplementedError(self.delete_handler)

    def copy_handler(self, filecmd):
        """Handle a filecopy command."""
        raise NotImplementedError(self.copy_handler)

    def rename_handler(self, filecmd):
        """Handle a filerename command."""
        raise NotImplementedError(self.rename_handler)

    def deleteall_handler(self, filecmd):
        """Handle a filedeleteall command."""
        raise NotImplementedError(self.deleteall_handler)



# Copyright (C) 2008-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Parser of import data into command objects.

In order to reuse existing front-ends, the stream format is a subset of
the one used by git-fast-import (as of the 1.5.4 release of git at least).
The grammar is:

  stream ::= cmd*;

  cmd ::= new_blob
        | new_commit
        | new_tag
        | reset_branch
        | checkpoint
        | progress
        ;

  new_blob ::= 'blob' lf
    mark?
    file_content;
  file_content ::= data;

  new_commit ::= 'commit' sp ref_str lf
    mark?
    ('author' sp name '<' email '>' when lf)?
    'committer' sp name '<' email '>' when lf
    commit_msg
    ('from' sp (ref_str | hexsha1 | sha1exp_str | idnum) lf)?
    ('merge' sp (ref_str | hexsha1 | sha1exp_str | idnum) lf)*
    file_change*
    lf?;
  commit_msg ::= data;

  file_change ::= file_clr
    | file_del
    | file_rnm
    | file_cpy
    | file_obm
    | file_inm;
  file_clr ::= 'deleteall' lf;
  file_del ::= 'D' sp path_str lf;
  file_rnm ::= 'R' sp path_str sp path_str lf;
  file_cpy ::= 'C' sp path_str sp path_str lf;
  file_obm ::= 'M' sp mode sp (hexsha1 | idnum) sp path_str lf;
  file_inm ::= 'M' sp mode sp 'inline' sp path_str lf
    data;

  new_tag ::= 'tag' sp tag_str lf
    'from' sp (ref_str | hexsha1 | sha1exp_str | idnum) lf
    'tagger' sp name '<' email '>' when lf
    tag_msg;
  tag_msg ::= data;

  reset_branch ::= 'reset' sp ref_str lf
    ('from' sp (ref_str | hexsha1 | sha1exp_str | idnum) lf)?
    lf?;

  checkpoint ::= 'checkpoint' lf
    lf?;

  progress ::= 'progress' sp not_lf* lf
    lf?;

     # note: the first idnum in a stream should be 1 and subsequent
     # idnums should not have gaps between values as this will cause
     # the stream parser to reserve space for the gapped values.  An
     # idnum can be updated in the future to a new object by issuing
     # a new mark directive with the old idnum.
     #
  mark ::= 'mark' sp idnum lf;
  data ::= (delimited_data | exact_data)
    lf?;

    # note: delim may be any string but must not contain lf.
    # data_line may contain any data but must not be exactly
    # delim. The lf after the final data_line is included in
    # the data.
  delimited_data ::= 'data' sp '<<' delim lf
    (data_line lf)*
    delim lf;

     # note: declen indicates the length of binary_data in bytes.
     # declen does not include the lf preceeding the binary data.
     #
  exact_data ::= 'data' sp declen lf
    binary_data;

     # note: quoted strings are C-style quoting supporting \\c for
     # common escapes of 'c' (e..g \\n, \\t, \\\\, \\") or \\nnn where nnn
     # is the signed byte value in octal.  Note that the only
     # characters which must actually be escaped to protect the
     # stream formatting is: \\, " and LF.  Otherwise these values
     # are UTF8.
     #
  ref_str     ::= ref;
  sha1exp_str ::= sha1exp;
  tag_str     ::= tag;
  path_str    ::= path    | '"' quoted(path)    '"' ;
  mode        ::= '100644' | '644'
                | '100755' | '755'
                | '120000'
                ;

  declen ::= # unsigned 32 bit value, ascii base10 notation;
  bigint ::= # unsigned integer value, ascii base10 notation;
  binary_data ::= # file content, not interpreted;

  when         ::= raw_when | rfc2822_when;
  raw_when     ::= ts sp tz;
  rfc2822_when ::= # Valid RFC 2822 date and time;

  sp ::= # ASCII space character;
  lf ::= # ASCII newline (LF) character;

     # note: a colon (':') must precede the numerical value assigned to
     # an idnum.  This is to distinguish it from a ref or tag name as
     # GIT does not permit ':' in ref or tag strings.
     #
  idnum   ::= ':' bigint;
  path    ::= # GIT style file path, e.g. "a/b/c";
  ref     ::= # GIT ref name, e.g. "refs/heads/MOZ_GECKO_EXPERIMENT";
  tag     ::= # GIT tag name, e.g. "FIREFOX_1_5";
  sha1exp ::= # Any valid GIT SHA1 expression;
  hexsha1 ::= # SHA1 in hexadecimal format;

     # note: name and email are UTF8 strings, however name must not
     # contain '<' or lf and email must not contain any of the
     # following: '<', '>', lf.
     #
  name  ::= # valid GIT author/committer name;
  email ::= # valid GIT author/committer email;
  ts    ::= # time since the epoch in seconds, ascii base10 notation;
  tz    ::= # GIT style timezone;

     # note: comments may appear anywhere in the input, except
     # within a data command.  Any form of the data command
     # always escapes the related input from comment processing.
     #
     # In case it is not clear, the '#' that starts the comment
     # must be the first character on that the line (an lf have
     # preceeded it).
     #
  comment ::= '#' not_lf* lf;
  not_lf  ::= # Any byte that is not ASCII newline (LF);
"""
from __future__ import print_function

import collections
import re
import sys
import codecs

from . import (
    commands,
    dates,
    errors,
    )
from .helpers import (
    newobject as object,
    utf8_bytes_string,
    )


class LineBasedParser(object):

    def __init__(self, input_stream):
        """A Parser that keeps track of line numbers.

        :param input: the file-like object to read from
        """
        self.input = input_stream
        self.lineno = 0
        # Lines pushed back onto the input stream
        self._buffer = []

    def abort(self, exception, *args):
        """Raise an exception providing line number information."""
        raise exception(self.lineno, *args)

    def readline(self):
        """Get the next line including the newline or '' on EOF."""
        self.lineno += 1
        if self._buffer:
            return self._buffer.pop()
        else:
            return self.input.readline()

    def next_line(self):
        """Get the next line without the newline or None on EOF."""
        line = self.readline()
        if line:
            return line[:-1]
        else:
            return None

    def push_line(self, line):
        """Push line back onto the line buffer.

        :param line: the line with no trailing newline
        """
        self.lineno -= 1
        self._buffer.append(line + b'\n')

    def read_bytes(self, count):
        """Read a given number of bytes from the input stream.

        Throws MissingBytes if the bytes are not found.

        Note: This method does not read from the line buffer.

        :return: a string
        """
        result = self.input.read(count)
        found = len(result)
        self.lineno += result.count(b'\n')
        if found != count:
            self.abort(errors.MissingBytes, count, found)
        return result

    def read_until(self, terminator):
        """Read the input stream until the terminator is found.

        Throws MissingTerminator if the terminator is not found.

        Note: This method does not read from the line buffer.

        :return: the bytes read up to but excluding the terminator.
        """

        lines = []
        term = terminator + b'\n'
        while True:
            line = self.input.readline()
            if line == term:
                break
            else:
                lines.append(line)
        return b''.join(lines)


# Regular expression used for parsing. (Note: The spec states that the name
# part should be non-empty but git-fast-export doesn't always do that so
# the first bit is \w*, not \w+.) Also git-fast-import code says the
# space before the email is optional.
_WHO_AND_WHEN_RE = re.compile(br'([^<]*)<(.*)> (.+)')
_WHO_RE = re.compile(br'([^<]*)<(.*)>')


class ImportParser(LineBasedParser):

    def __init__(self, input_stream, verbose=False, output=sys.stdout,
                 user_mapper=None, strict=True):
        """A Parser of import commands.

        :param input_stream: the file-like object to read from
        :param verbose: display extra information of not
        :param output: the file-like object to write messages to (YAGNI?)
        :param user_mapper: if not None, the UserMapper used to adjust
          user-ids for authors, committers and taggers.
        :param strict: Raise errors on strictly invalid data
        """
        LineBasedParser.__init__(self, input_stream)
        self.verbose = verbose
        self.output = output
        self.user_mapper = user_mapper
        self.strict = strict
        # We auto-detect the date format when a date is first encountered
        self.date_parser = None
        self.features = {}

    def warning(self, msg):
        sys.stderr.write("warning line %d: %s\n" % (self.lineno, msg))

    def iter_commands(self):
        """Iterator returning ImportCommand objects."""
        while True:
            line = self.next_line()
            if line is None:
                if b'done' in self.features:
                    raise errors.PrematureEndOfStream(self.lineno)
                break
            elif len(line) == 0 or line.startswith(b'#'):
                continue
            # Search for commands in order of likelihood
            elif line.startswith(b'commit '):
                yield self._parse_commit(line[len(b'commit '):])
            elif line.startswith(b'blob'):
                yield self._parse_blob()
            elif line.startswith(b'done'):
                break
            elif line.startswith(b'progress '):
                yield commands.ProgressCommand(line[len(b'progress '):])
            elif line.startswith(b'reset '):
                yield self._parse_reset(line[len(b'reset '):])
            elif line.startswith(b'tag '):
                yield self._parse_tag(line[len(b'tag '):])
            elif line.startswith(b'checkpoint'):
                yield commands.CheckpointCommand()
            elif line.startswith(b'feature'):
                yield self._parse_feature(line[len(b'feature '):])
            else:
                self.abort(errors.InvalidCommand, line)

    def iter_file_commands(self):
        """Iterator returning FileCommand objects.

        If an invalid file command is found, the line is silently
        pushed back and iteration ends.
        """
        while True:
            line = self.next_line()
            if line is None:
                break
            elif len(line) == 0 or line.startswith(b'#'):
                continue
            # Search for file commands in order of likelihood
            elif line.startswith(b'M '):
                yield self._parse_file_modify(line[2:])
            elif line.startswith(b'D '):
                path = self._path(line[2:])
                yield commands.FileDeleteCommand(path)
            elif line.startswith(b'R '):
                old, new = self._path_pair(line[2:])
                yield commands.FileRenameCommand(old, new)
            elif line.startswith(b'C '):
                src, dest = self._path_pair(line[2:])
                yield commands.FileCopyCommand(src, dest)
            elif line.startswith(b'deleteall'):
                yield commands.FileDeleteAllCommand()
            else:
                self.push_line(line)
                break

    def _parse_blob(self):
        """Parse a blob command."""
        lineno = self.lineno
        mark = self._get_mark_if_any()
        data = self._get_data(b'blob')
        return commands.BlobCommand(mark, data, lineno)

    def _parse_commit(self, ref):
        """Parse a commit command."""
        lineno = self.lineno
        mark = self._get_mark_if_any()
        author = self._get_user_info(b'commit', b'author', False)
        more_authors = []
        while True:
            another_author = self._get_user_info(b'commit', b'author', False)
            if another_author is not None:
                more_authors.append(another_author)
            else:
                break
        committer = self._get_user_info(b'commit', b'committer')
        message = self._get_data(b'commit', b'message')
        from_ = self._get_from()
        merges = []
        while True:
            merge = self._get_merge()
            if merge is not None:
                # while the spec suggests it's illegal, git-fast-export
                # outputs multiple merges on the one line, e.g.
                # merge :x :y :z
                these_merges = merge.split(b' ')
                merges.extend(these_merges)
            else:
                break
        properties = {}
        while True:
            name_value = self._get_property()
            if name_value is not None:
                name, value = name_value
                properties[name] = value
            else:
                break
        return commands.CommitCommand(
            ref, mark, author, committer, message,
            from_, merges, list(self.iter_file_commands()), lineno=lineno,
            more_authors=more_authors, properties=properties)

    def _parse_feature(self, info):
        """Parse a feature command."""
        parts = info.split(b'=', 1)
        name = parts[0]
        if len(parts) > 1:
            value = self._path(parts[1])
        else:
            value = None
        self.features[name] = value
        return commands.FeatureCommand(name, value, lineno=self.lineno)

    def _parse_file_modify(self, info):
        """Parse a filemodify command within a commit.

        :param info: a string in the format "mode dataref path"
          (where dataref might be the hard-coded literal 'inline').
        """
        params = info.split(b' ', 2)
        path = self._path(params[2])
        mode = self._mode(params[0])
        if params[1] == b'inline':
            dataref = None
            data = self._get_data(b'filemodify')
        else:
            dataref = params[1]
            data = None
        return commands.FileModifyCommand(
            path, mode, dataref, data)

    def _parse_reset(self, ref):
        """Parse a reset command."""
        from_ = self._get_from()
        return commands.ResetCommand(ref, from_)

    def _parse_tag(self, name):
        """Parse a tag command."""
        from_ = self._get_from(b'tag')
        tagger = self._get_user_info(
            b'tag', b'tagger', accept_just_who=True)
        message = self._get_data(b'tag', b'message')
        return commands.TagCommand(name, from_, tagger, message)

    def _get_mark_if_any(self):
        """Parse a mark section."""
        line = self.next_line()
        if line.startswith(b'mark :'):
            return line[len(b'mark :'):]
        else:
            self.push_line(line)
            return None

    def _get_from(self, required_for=None):
        """Parse a from section."""
        line = self.next_line()
        if line is None:
            return None
        elif line.startswith(b'from '):
            return line[len(b'from '):]
        elif required_for:
            self.abort(errors.MissingSection, required_for, 'from')
        else:
            self.push_line(line)
            return None

    def _get_merge(self):
        """Parse a merge section."""
        line = self.next_line()
        if line is None:
            return None
        elif line.startswith(b'merge '):
            return line[len(b'merge '):]
        else:
            self.push_line(line)
            return None

    def _get_property(self):
        """Parse a property section."""
        line = self.next_line()
        if line is None:
            return None
        elif line.startswith(b'property '):
            return self._name_value(line[len(b'property '):])
        else:
            self.push_line(line)
            return None

    def _get_user_info(self, cmd, section, required=True,
                       accept_just_who=False):
        """Parse a user section."""
        line = self.next_line()
        if line.startswith(section + b' '):
            return self._who_when(
                line[len(section + b' '):], cmd, section,
                accept_just_who=accept_just_who)
        elif required:
            self.abort(errors.MissingSection, cmd, section)
        else:
            self.push_line(line)
            return None

    def _get_data(self, required_for, section=b'data'):
        """Parse a data section."""
        line = self.next_line()
        if line.startswith(b'data '):
            rest = line[len(b'data '):]
            if rest.startswith(b'<<'):
                return self.read_until(rest[2:])
            else:
                size = int(rest)
                read_bytes = self.read_bytes(size)
                # optional LF after data.
                next_line = self.input.readline()
                self.lineno += 1
                if len(next_line) > 1 or next_line != b'\n':
                    self.push_line(next_line[:-1])
                return read_bytes
        else:
            self.abort(errors.MissingSection, required_for, section)

    def _who_when(self, s, cmd, section, accept_just_who=False):
        """Parse who and when information from a string.

        :return: a tuple of (name,email,timestamp,timezone). name may be
            the empty string if only an email address was given.
        """
        match = _WHO_AND_WHEN_RE.search(s)
        if match:
            datestr = match.group(3).lstrip()
            if self.date_parser is None:
                # auto-detect the date format
                if len(datestr.split(b' ')) == 2:
                    date_format = 'raw'
                elif datestr == b'now':
                    date_format = 'now'
                else:
                    date_format = 'rfc2822'
                self.date_parser = dates.DATE_PARSERS_BY_NAME[date_format]
            try:
                when = self.date_parser(datestr, self.lineno)
            except ValueError:
                print("failed to parse datestr '%s'" % (datestr,))
                raise
            name = match.group(1).rstrip()
            email = match.group(2)
        else:
            match = _WHO_RE.search(s)
            if accept_just_who and match:
                # HACK around missing time
                # TODO: output a warning here
                when = dates.DATE_PARSERS_BY_NAME['now']('now')
                name = match.group(1)
                email = match.group(2)
            elif self.strict:
                self.abort(errors.BadFormat, cmd, section, s)
            else:
                name = s
                email = None
                when = dates.DATE_PARSERS_BY_NAME['now']('now')
        if len(name) > 0:
            if name.endswith(b' '):
                name = name[:-1]
        # While it shouldn't happen, some datasets have email addresses
        # which contain unicode characters. See bug 338186. We sanitize
        # the data at this level just in case.
        if self.user_mapper:
            name, email = self.user_mapper.map_name_and_email(name, email)

        return Authorship(name, email, when[0], when[1])

    def _name_value(self, s):
        """Parse a (name,value) tuple from 'name value-length value'."""
        parts = s.split(b' ', 2)
        name = parts[0]
        if len(parts) == 1:
            value = None
        else:
            size = int(parts[1])
            value = parts[2]
            still_to_read = size - len(value)
            if still_to_read > 0:
                read_bytes = self.read_bytes(still_to_read)
                value += b'\n' + read_bytes[:still_to_read - 1]
        return (name, value)

    def _path(self, s):
        """Parse a path."""
        if s.startswith(b'"'):
            if not s.endswith(b'"'):
                self.abort(errors.BadFormat, '?', '?', s)
            else:
                return _unquote_c_string(s[1:-1])
        return s

    def _path_pair(self, s):
        """Parse two paths separated by a space."""
        # TODO: handle a space in the first path
        if s.startswith(b'"'):
            parts = s[1:].split(b'" ', 1)
        else:
            parts = s.split(b' ', 1)
        if len(parts) != 2:
            self.abort(errors.BadFormat, '?', '?', s)
        elif parts[1].startswith(b'"') and parts[1].endswith(b'"'):
            parts[1] = parts[1][1:-1]
        elif parts[1].startswith(b'"') or parts[1].endswith(b'"'):
            self.abort(errors.BadFormat, '?', '?', s)
        return [_unquote_c_string(part) for part in parts]

    def _mode(self, s):
        """Check file mode format and parse into an int.

        :return: mode as integer
        """
        # Note: Output from git-fast-export slightly different to spec
        if s in [b'644', b'100644', b'0100644']:
            return 0o100644
        elif s in [b'755', b'100755', b'0100755']:
            return 0o100755
        elif s in [b'040000', b'0040000']:
            return 0o40000
        elif s in [b'120000', b'0120000']:
            return 0o120000
        elif s in [b'160000', b'0160000']:
            return 0o160000
        else:
            self.abort(errors.BadFormat, 'filemodify', 'mode', s)


ESCAPE_SEQUENCE_BYTES_RE = re.compile(br'''
    ( \\U........      # 8-digit hex escapes
    | \\u....          # 4-digit hex escapes
    | \\x..            # 2-digit hex escapes
    | \\[0-7]{1,3}     # Octal escapes
    | \\N\{[^}]+\}     # Unicode characters by name
    | \\[\\'"abfnrtv]  # Single-character escapes
    )''', re.VERBOSE)


ESCAPE_SEQUENCE_RE = re.compile(r'''
    ( \\U........
    | \\u....
    | \\x..
    | \\[0-7]{1,3}
    | \\N\{[^}]+\}
    | \\[\\'"abfnrtv]
    )''', re.UNICODE | re.VERBOSE)


def _unquote_c_string(s):
    """replace C-style escape sequences (\n, \", etc.) with real chars."""
    # doing a s.encode('utf-8').decode('unicode_escape') can return an
    # incorrect output with unicode string (both in py2 and py3) the safest way
    # is to match the escape sequences and decoding them alone.
    def decode_match(match):
        return utf8_bytes_string(
            codecs.decode(match.group(0), 'unicode-escape')
        )

    if isinstance(s, bytes):
        return ESCAPE_SEQUENCE_BYTES_RE.sub(decode_match, s)
    else:
        return ESCAPE_SEQUENCE_RE.sub(decode_match, s)


Authorship = collections.namedtuple(
    'Authorship', 'name email timestamp timezone')



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Import processor that queries the input (and doesn't import)."""
from __future__ import print_function


from .. import (
    commands,
    processor,
    )


class QueryProcessor(processor.ImportProcessor):
    """An import processor that queries the input.

    No changes to the current repository are made.
    """

    known_params = (
        commands.COMMAND_NAMES +
        commands.FILE_COMMAND_NAMES +
        [b'commit-mark']
    )

    def __init__(self, params=None, verbose=False):
        processor.ImportProcessor.__init__(self, params, verbose)
        self.parsed_params = {}
        self.interesting_commit = None
        self._finished = False
        if params:
            if 'commit-mark' in params:
                self.interesting_commit = params['commit-mark']
                del params['commit-mark']
            for name, value in params.items():
                if value == 1:
                    # All fields
                    fields = None
                else:
                    fields = value.split(',')
                self.parsed_params[name] = fields

    def pre_handler(self, cmd):
        """Hook for logic before each handler starts."""
        if self._finished:
            return
        if self.interesting_commit and cmd.name == 'commit':
            if cmd.mark == self.interesting_commit:
                print(cmd.to_string())
                self._finished = True
            return
        if cmd.name in self.parsed_params:
            fields = self.parsed_params[cmd.name]
            str = cmd.dump_str(fields, self.parsed_params, self.verbose)
            print("%s" % (str,))

    def progress_handler(self, cmd):
        """Process a ProgressCommand."""
        pass

    def blob_handler(self, cmd):
        """Process a BlobCommand."""
        pass

    def checkpoint_handler(self, cmd):
        """Process a CheckpointCommand."""
        pass

    def commit_handler(self, cmd):
        """Process a CommitCommand."""
        pass

    def reset_handler(self, cmd):
        """Process a ResetCommand."""
        pass

    def tag_handler(self, cmd):
        """Process a TagCommand."""
        pass

    def feature_handler(self, cmd):
        """Process a FeatureCommand."""
        feature = cmd.feature_name
        if feature not in commands.FEATURE_NAMES:
            self.warning(
                "feature %s is not supported - parsing may fail"
                % (feature,))



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the Import parsing"""
import io
import time
import unittest

from fastimport import (
    commands,
    errors,
    parser,
    )


class TestLineBasedParser(unittest.TestCase):

    def test_push_line(self):
        s = io.BytesIO(b"foo\nbar\nbaz\n")
        p = parser.LineBasedParser(s)
        self.assertEqual(b'foo', p.next_line())
        self.assertEqual(b'bar', p.next_line())
        p.push_line(b'bar')
        self.assertEqual(b'bar', p.next_line())
        self.assertEqual(b'baz', p.next_line())
        self.assertEqual(None, p.next_line())

    def test_read_bytes(self):
        s = io.BytesIO(b"foo\nbar\nbaz\n")
        p = parser.LineBasedParser(s)
        self.assertEqual(b'fo', p.read_bytes(2))
        self.assertEqual(b'o\nb', p.read_bytes(3))
        self.assertEqual(b'ar', p.next_line())
        # Test that the line buffer is ignored
        p.push_line(b'bar')
        self.assertEqual(b'baz', p.read_bytes(3))
        # Test missing bytes
        self.assertRaises(errors.MissingBytes, p.read_bytes, 10)

    def test_read_until(self):
        # TODO
        return
        s = io.BytesIO(b"foo\nbar\nbaz\nabc\ndef\nghi\n")
        p = parser.LineBasedParser(s)
        self.assertEqual(b'foo\nbar', p.read_until(b'baz'))
        self.assertEqual(b'abc', p.next_line())
        # Test that the line buffer is ignored
        p.push_line(b'abc')
        self.assertEqual(b'def', p.read_until(b'ghi'))
        # Test missing terminator
        self.assertRaises(errors.MissingTerminator, p.read_until(b'>>>'))


# Sample text
_sample_import_text = b"""
progress completed
# Test blob formats
blob
mark :1
data 4
aaaablob
data 5
bbbbb
# Commit formats
commit refs/heads/master
mark :2
committer bugs bunny <bugs@bunny.org> now
data 14
initial import
M 644 inline README
data 18
Welcome from bugs
commit refs/heads/master
committer <bugs@bunny.org> now
data 13
second commit
from :2
M 644 inline README
data 23
Welcome from bugs, etc.
# Miscellaneous
checkpoint
progress completed
# Test a commit without sub-commands (bug #351717)
commit refs/heads/master
mark :3
author <bugs@bunny.org> now
committer <bugs@bunny.org> now
data 20
first commit, empty
# Test a commit with a heredoc-style (delimited_data) messsage (bug #400960)
commit refs/heads/master
mark :4
author <bugs@bunny.org> now
committer <bugs@bunny.org> now
data <<EOF
Commit with heredoc-style message
EOF
# Test a "submodule"/tree-reference
commit refs/heads/master
mark :5
author <bugs@bunny.org> now
committer <bugs@bunny.org> now
data 15
submodule test
M 160000 rev-id tree-id
# Test features
feature whatever
feature foo=bar
# Test commit with properties
commit refs/heads/master
mark :6
committer <bugs@bunny.org> now
data 18
test of properties
property p1
property p2 5 hohum
property p3 16 alpha
beta
gamma
property p4 8 whatever
# Test a commit with multiple authors
commit refs/heads/master
mark :7
author Fluffy <fluffy@bunny.org> now
author Daffy <daffy@duck.org> now
author Donald <donald@duck.org> now
committer <bugs@bunny.org> now
data 17
multi-author test
"""

_timefunc = time.time


class TestImportParser(unittest.TestCase):

    def setUp(self):
        self.fake_time = 42.0123
        time.time = lambda: self.fake_time

    def tearDown(self):
        time.time = _timefunc
        del self.fake_time

    def test_iter_commands(self):
        s = io.BytesIO(_sample_import_text)
        p = parser.ImportParser(s)
        result = []
        for cmd in p.iter_commands():
            result.append(cmd)
            if cmd.name == b'commit':
                for fc in cmd.iter_files():
                    result.append(fc)

        self.assertEqual(len(result), 17)
        cmd1 = result.pop(0)
        self.assertEqual(b'progress', cmd1.name)
        self.assertEqual(b'completed', cmd1.message)
        cmd2 = result.pop(0)
        self.assertEqual(b'blob', cmd2.name)
        self.assertEqual(b'1', cmd2.mark)
        self.assertEqual(b':1', cmd2.id)
        self.assertEqual(b'aaaa', cmd2.data)
        self.assertEqual(4, cmd2.lineno)
        cmd3 = result.pop(0)
        self.assertEqual(b'blob', cmd3.name)
        self.assertEqual(b'@7', cmd3.id)
        self.assertEqual(None, cmd3.mark)
        self.assertEqual(b'bbbbb', cmd3.data)
        self.assertEqual(7, cmd3.lineno)
        cmd4 = result.pop(0)
        self.assertEqual(b'commit', cmd4.name)
        self.assertEqual(b'2', cmd4.mark)
        self.assertEqual(b':2', cmd4.id)
        self.assertEqual(
            b'initial import', cmd4.message)

        self.assertEqual(
            (b'bugs bunny', b'bugs@bunny.org', self.fake_time, 0),
            cmd4.committer)
        # namedtuple attributes
        self.assertEqual(b'bugs bunny', cmd4.committer.name)
        self.assertEqual(b'bugs@bunny.org', cmd4.committer.email)
        self.assertEqual(self.fake_time, cmd4.committer.timestamp)
        self.assertEqual(0, cmd4.committer.timezone)

        self.assertEqual(None, cmd4.author)
        self.assertEqual(11, cmd4.lineno)
        self.assertEqual(b'refs/heads/master', cmd4.ref)
        self.assertEqual(None, cmd4.from_)
        self.assertEqual([], cmd4.merges)
        file_cmd1 = result.pop(0)
        self.assertEqual(b'filemodify', file_cmd1.name)
        self.assertEqual(b'README', file_cmd1.path)
        self.assertEqual(0o100644, file_cmd1.mode)
        self.assertEqual(b'Welcome from bugs\n', file_cmd1.data)
        cmd5 = result.pop(0)
        self.assertEqual(b'commit', cmd5.name)
        self.assertEqual(None, cmd5.mark)
        self.assertEqual(b'@19', cmd5.id)
        self.assertEqual(b'second commit', cmd5.message)
        self.assertEqual(
            (b'', b'bugs@bunny.org', self.fake_time, 0), cmd5.committer)
        self.assertEqual(None, cmd5.author)
        self.assertEqual(19, cmd5.lineno)
        self.assertEqual(b'refs/heads/master', cmd5.ref)
        self.assertEqual(b':2', cmd5.from_)
        self.assertEqual([], cmd5.merges)
        file_cmd2 = result.pop(0)
        self.assertEqual(b'filemodify', file_cmd2.name)
        self.assertEqual(b'README', file_cmd2.path)
        self.assertEqual(0o100644, file_cmd2.mode)
        self.assertEqual(b'Welcome from bugs, etc.', file_cmd2.data)
        cmd6 = result.pop(0)
        self.assertEqual(cmd6.name, b'checkpoint')
        cmd7 = result.pop(0)
        self.assertEqual(b'progress', cmd7.name)
        self.assertEqual(b'completed', cmd7.message)
        cmd = result.pop(0)
        self.assertEqual(b'commit', cmd.name)
        self.assertEqual(b'3', cmd.mark)
        self.assertEqual(None, cmd.from_)
        cmd = result.pop(0)
        self.assertEqual(b'commit', cmd.name)
        self.assertEqual(b'4', cmd.mark)
        self.assertEqual(b'Commit with heredoc-style message\n', cmd.message)
        cmd = result.pop(0)
        self.assertEqual(b'commit', cmd.name)
        self.assertEqual(b'5', cmd.mark)
        self.assertEqual(b'submodule test\n', cmd.message)
        file_cmd1 = result.pop(0)
        self.assertEqual(b'filemodify', file_cmd1.name)
        self.assertEqual(b'tree-id', file_cmd1.path)
        self.assertEqual(0o160000, file_cmd1.mode)
        self.assertEqual(b"rev-id", file_cmd1.dataref)
        cmd = result.pop(0)
        self.assertEqual(b'feature', cmd.name)
        self.assertEqual(b'whatever', cmd.feature_name)
        self.assertEqual(None, cmd.value)
        cmd = result.pop(0)
        self.assertEqual(b'feature', cmd.name)
        self.assertEqual(b'foo', cmd.feature_name)
        self.assertEqual(b'bar', cmd.value)
        cmd = result.pop(0)
        self.assertEqual(b'commit', cmd.name)
        self.assertEqual(b'6', cmd.mark)
        self.assertEqual(b'test of properties', cmd.message)
        self.assertEqual({
            b'p1': None,
            b'p2': b'hohum',
            b'p3': b'alpha\nbeta\ngamma',
            b'p4': b'whatever',
        }, cmd.properties)
        cmd = result.pop(0)
        self.assertEqual(b'commit', cmd.name)
        self.assertEqual(b'7', cmd.mark)
        self.assertEqual(b'multi-author test', cmd.message)
        self.assertEqual(b'', cmd.committer[0])
        self.assertEqual(b'bugs@bunny.org', cmd.committer[1])
        self.assertEqual(b'Fluffy', cmd.author[0])
        self.assertEqual(b'fluffy@bunny.org', cmd.author[1])
        self.assertEqual(b'Daffy', cmd.more_authors[0][0])
        self.assertEqual(b'daffy@duck.org', cmd.more_authors[0][1])
        self.assertEqual(b'Donald', cmd.more_authors[1][0])
        self.assertEqual(b'donald@duck.org', cmd.more_authors[1][1])

    def test_done_feature_missing_done(self):
        s = io.BytesIO(b"""feature done
""")
        p = parser.ImportParser(s)
        cmds = p.iter_commands()
        self.assertEqual(b"feature", next(cmds).name)
        self.assertRaises(errors.PrematureEndOfStream, lambda: next(cmds))

    def test_done_with_feature(self):
        s = io.BytesIO(b"""feature done
done
more data
""")
        p = parser.ImportParser(s)
        cmds = p.iter_commands()
        self.assertEqual(b"feature", next(cmds).name)
        self.assertRaises(StopIteration, lambda: next(cmds))

    def test_done_without_feature(self):
        s = io.BytesIO(b"""done
more data
""")
        p = parser.ImportParser(s)
        cmds = p.iter_commands()
        self.assertEqual([], list(cmds))


class TestStringParsing(unittest.TestCase):

    def test_unquote(self):
        s = br'hello \"sweet\" wo\\r\tld'
        self.assertEqual(
            br'hello "sweet" wo\r' + b'\tld',
            parser._unquote_c_string(s))


class TestPathPairParsing(unittest.TestCase):

    def test_path_pair_simple(self):
        p = parser.ImportParser(b'')
        self.assertEqual([b'foo', b'bar'], p._path_pair(b'foo bar'))

    def test_path_pair_spaces_in_first(self):
        p = parser.ImportParser("")
        self.assertEqual(
            [b'foo bar', b'baz'],
            p._path_pair(b'"foo bar" baz'))


class TestTagParsing(unittest.TestCase):

    def test_tagger_with_email(self):
        p = parser.ImportParser(io.BytesIO(
            b"tag refs/tags/v1.0\n"
            b"from :xxx\n"
            b"tagger Joe Wong <joe@example.com> 1234567890 -0600\n"
            b"data 11\n"
            b"create v1.0"))
        cmds = list(p.iter_commands())
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], commands.TagCommand))
        self.assertEqual(
            cmds[0].tagger,
            (b'Joe Wong', b'joe@example.com', 1234567890.0, -21600))

    def test_tagger_no_email_strict(self):
        p = parser.ImportParser(io.BytesIO(
            b"tag refs/tags/v1.0\n"
            b"from :xxx\n"
            b"tagger Joe Wong\n"
            b"data 11\n"
            b"create v1.0"))
        self.assertRaises(errors.BadFormat, list, p.iter_commands())

    def test_tagger_no_email_not_strict(self):
        p = parser.ImportParser(io.BytesIO(
            b"tag refs/tags/v1.0\n"
            b"from :xxx\n"
            b"tagger Joe Wong\n"
            b"data 11\n"
            b"create v1.0"), strict=False)
        cmds = list(p.iter_commands())
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], commands.TagCommand))
        self.assertEqual(cmds[0].tagger[:2], (b'Joe Wong', None))
